import sys
from os import mkdir, listdir
from decimal import Inexact


ROUTES_TEXT: str = \
"""from mizuhara.core.routes import (connector_callback,
                                  connector_command,
                                  connector_message,)


# Mapping handlers and views for Telegram Bot commands.
COMMANDS: list = [
]


MESSAGES: list = [
]


CALLBACKS: list = [
]

"""

SERIALIZERS_TEXT: str = \
"""# please import handlers classes in core.handlers
# and inherit one of them to your custom handler serializer classes.
from mizuhara.core.handlers.handlers import *
from mizuhara.core.routes import CLIENT_INFO


# please write down your code below.

"""

VIEWS_TEXT: str = \
"""# please write down your code below.

"""

EXECUTE_TEXT: str = \
"""import asyncio
from importlib import import_module
from telebot.async_telebot import (logger,
                                   AsyncTeleBot,)
from config import *


# create a Telegram Bot instance
bot = AsyncTeleBot(token=TELEBOT_TOKEN,
                   parse_mode=PARSE_MODE,
                   offset=OFFSET,
                   exception_handler=EXCEPTION_HANDLER,
                   state_storage=STATE_STORAGE,
                   disable_web_page_preview=DISABLE_WEB_PAGE_PREVIEW,
                   disable_notification=DISABLE_NOTIFICATION,
                   protect_content=PROTECT_CONTENT,
                   allow_sending_without_reply=ALLOW_SENDING_WITHOUT_REPLY,
                   colorful_logs=COLORFUL_LOG,
                   validate_token=VALIDATE_TOKEN)


# set Log Level
logger.setLevel(level=LOG_LEVEL)


# main for set menu and polling.
async def execute() -> None:
    await bot.set_my_commands(commands=MENU_COMMANDS)
    await bot.polling()
    return None


# start main
if __name__ == "__main__":
    # import handlers in installed app.
    for app_name in INSTALLED_APPS:
        try:
            module = import_module(f"{app_name}.routes")
            names_to_import = [name for name in dir(module) if not name.startswith("_")]
            for name in names_to_import:
                globals()[name] = getattr(module, name)

        except (ImportError, AttributeError) as e:
            if isinstance(e, ImportError):
                logger.error(f"Improper App Name '{app_name}': {e}")

            else:
                logger.error(f"accessing attributes in module {app_name}: {e}")

            break

    # execute polling in asynchronous
    asyncio.run(execute())
"""

CONFIG_TEXT: str = \
"""# Set the argument for asynchronous telebot instance.

from os import getenv
from telebot.types import BotCommand


### TELEGRAM BOT CONFIG ###
# Set the Telegram Bot API TOKEN for Telegram Bot Father
# User must set this value before executing main.py
# it is recommended to export the value for environment variables.
# - sh: export TELEBOT_TOKEN=YOUR_BOT_TOKEN
TELEBOT_TOKEN: str = getenv("TELEBOT_TOKEN")


# Set the parse mode for Telegram Bot.
# default is None.
# possible values: "HTML", "Markdown"
PARSE_MODE: str | None = None


# Set the offset for Telegram Bot.
# default is None.
# possible value: positive int
# process the requests of message after specific ID' message
OFFSET: int | None = None


# Set the exception for Telegram Bot.
# import custom exception handler and assign it to EXCEPTION_HANDLER.
# default is handlers.exception_handlers.custom_exception_handler
EXCEPTION_HANDLER = None


# Set the way to store and maintain the status of Telegram Bot.
# bot will read the last status of telegram bot from file, DB or memory during restart process.
# default is None and it will store the status information in memory.
# possible values: class in telebot.storage package
# - StateMemoryStorage, StateStorageBase(use class inheriting this class), StateRedisStorage
STATE_STORAGE = None


# Set the web_page_preview config.
# it will show a preview of a linked web page(protocol://host.domain) in message or not
# default is None(False)
DISABLE_WEB_PAGE_PREVIEW: bool | None = False


# Set the notification for bot message(alarm or vibration on your phone).
# default is None(False) - will give you alarm or vibration.
DISABLE_NOTIFICATION: bool | None = False


# Set the message protection
# preventing users from copying and forwarding sent bot message.
# default is None(False)
PROTECT_CONTENT: bool | None = True


# Set the bot's action about sending message to user.
# bot can only send standalone message after user's request if the value is set for False
# default is None(True)
ALLOW_SENDING_WITHOUT_REPLY: bool | None = False


# Set the colorful log config
# require 'coloredlogs' package in pip before using it.
# default is None(False)
COLORFUL_LOG: bool | None = True


# Set whether the bot validates its token or not.
# default is True
VALIDATE_TOKEN: bool | None = True


# Import handlers
INSTALLED_APPS: list | tuple = (
)

### TELEGRAM BOT CONFIG END ###

### TELEGRAM BOT FRAMEWORK CONFIG ###
# Set the allowed chat_type.
# default is ["private"]
# possible values: "private", "group", "supergroup", "channel"
ALLOWED_CHAT_TYPE: list = ["private"]


# Set menu button for your bot.
# use BotCommand(command="YOUR_COMMAND", description="COMMAND_DESCRIPTION")
# command name should be lower cases.
# if you change MENU_COMMANDS, remove your bot and re-enter to apply changes.
MENU_COMMANDS: list = [
    BotCommand(command="main", description="Main"),
]

# Set chatting mode for your bot.
# SECRET_MODE is a bool variable that decides whether the previous messages will be removed or not.
# if True, bot will remove all remained messages in chat room after user types messages or click InlineMarkupButton
SECRET_MODE: bool = True

### TELEGRAM BOT FRAMEWORK CONFIG END###

### LOGGING CONFIG ###
# Set the log level
# default is ERROR
# possible values: "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
LOG_LEVEL = "ERROR"


### LOGGING CONFIG END ###

"""


def main():
    """
    code for CLI command that create a telebot_framework project.

    :return: None
    """

    argv = sys.argv[1:]

    try:
        subcommand = argv[0]
        if subcommand == "newproject":
            create_project()

        elif subcommand == "newapp":
            create_app(name=argv[1])

        return None

    except IndexError as e:
        pass

    print_help()
    return None


def create_project():
    # create file: config.py
    with open("./config.py", mode="w", encoding="utf-8") as f:
        f.write(CONFIG_TEXT)

    # create file: execute.py
    with open("./execute.py", mode="w", encoding="utf-8") as f:
        f.write(EXECUTE_TEXT)

    return None

def create_app(name:str):
    # check the project initiation
    if ["config.py", "execute.py"] in listdir():
        create_project()

    mkdir(f"./{name}")

    # create file: __init__.py
    with open(f"./{name}/__init__.py", mode="w", encoding="utf-8") as f:
        f.write("")

    # create file: routes.py
    with open(f"./{name}/routes.py", mode="w", encoding="utf-8") as f:
        f.write(ROUTES_TEXT)

    with open(f"./{name}/serializers.py", mode="w", encoding="utf-8") as f:
        f.write(SERIALIZERS_TEXT)

    with open(f"./{name}/views.py", mode="w", encoding="utf-8") as f:
        f.write(VIEWS_TEXT)

    return None


def print_help():
    """
    code for printing out usage of command 'mizuhara'

    :return: None
    """

    print("Help Message")
    pass



from urllib.parse import urlencode
from requests import get as requests_get
from telebot.types import ForceReply
from mizuhara.core.handlers.handlers import ResultShowingWithInlineMarkup
from mizuhara.translation import translate


class ReceiverWithLocation(ResultShowingWithInlineMarkup):
    """
    ReceiverWithLocation:

    this class is responsible for sending location of telegram user.
    when user sends its location on Attachment > location, telegram can grap the location information(latitude and longitude).

    this class can be used to update current location or record travel routes and so on.
    user must send its location manually.
    """

    def __init__(self, types, **kwargs):
        super(ReceiverWithLocation, self).__init__(types, **kwargs)
        self.location = None


    async def get_location(self) -> bool:
        """
        this method is charge of sending preamble message and receiving user's current location.
        you can reprocess the location information with self.location:dict

        :return: bool
        """

        if getattr(self.types, "location", None) is not None:
            self.location = self.types.location.to_dict()
            await self.post_process()

            await self.bot.send_message(chat_id=self.chat_id,
                                        text=self.bot_text)
            return True

        await super().send_message()
        return False


class SenderWithLocation(ResultShowingWithInlineMarkup):
    """
    SenderWithLocation:

    this class is charge of sending location information with latitude and longitude.
    the dev must set the self.latitude and self.longitude by overriding pre_process.
    """

    def __init__(self, types, **kwargs):
        super(SenderWithLocation, self).__init__(types, **kwargs)
        self.latitude = None
        self.longitude = None

    async def pre_process(self):
        """
        by overriding this method, set the latitude and longitude (float).

        :return:
        """
        pass

    async def send_message(self):
        await self.pre_process()

        if self.latitude is None or self.longitude is None:
            raise ValueError(translate(domain="default_warnings",
                                       key="warn_location_not_found",
                                       language_code=self.language))

        if not isinstance(self.latitude, float) or not isinstance(self.longitude, float):
            raise ValueError(translate(domain="default_exceptions",
                                       key="err_location_wrong_type",
                                       language_code=self.language))

        await self.bot.send_location(chat_id=self.chat_id,
                                     latitude=self.latitude,
                                     longitude=self.longitude,
                                     reply_markup=self.bot_markup)
        return None


class SendWithLocationName(ResultShowingWithInlineMarkup):
    """
    SendWithLocationName

    this class is charge of searching location name, address or postal number, which is typed by telegram user.
    this class will return location data with longitude and latitude to telegram user.
    this class contains 3rd API service(Nominatim OpenStreetMap)

    """

    GEO_INFO_API_URL: str = "https://nominatim.openstreetmap.org/search?"

    def __init__(self, types, **kwargs):
        super(SendWithLocationName, self).__init__(types, **kwargs)
        self.latitude = None
        self.longitude = None

    async def send_message(self) -> None:
        # if user input the search query.
        if getattr(self.types, "message", None) is None:
            await self._remove_prev_message()
            await self.post_process()

        # initial stage that the user click the search location callback data.
        else:
            self.bot_text = translate(domain="default_handlers",
                                      key="guide_force_reply_continue",
                                      language_code=self.language)
            await super().send_message()
            self.bot_markup = ForceReply()
            await self.bot.send_message(chat_id=self.chat_id,
                                        text=translate(domain="default_handlers",
                                                       key="guide_location_input",
                                                       language_code=self.language),
                                        reply_markup=ForceReply())
        return None

    async def post_process(self) -> None:
        """
        this method calls Nominatim API and provide user to location information.

        :return:
        """

        headers = {"User-Agent": "telebot_framework"}
        params = {
            "q": self.client_response,
            "format": "jsonv2",
            "limit": 1
        }
        response = requests_get(url=self.GEO_INFO_API_URL + urlencode(params),
                                headers=headers)

        if response.status_code == 200 and len(response.json()) != 0:
            location = response.json()[0]
            address = location.get("display_name")
            self.latitude = location.get("lat")
            self.longitude = location.get("lon")
            text = translate(domain="default_handlers",
                             key="result_location_search",
                             language_code=self.language).format(self.client_response, address)

            await self.bot.send_message(chat_id=self.chat_id,
                                        text=text)

            await self.bot.send_location(chat_id=self.chat_id,
                                         latitude=self.latitude,
                                         longitude=self.longitude,
                                         reply_markup=self.bot_markup)
            return None

        else:
            self.bot_text = translate(domain="default_warnings",
                                       key="warn_location_not_found",
                                       language_code=self.language)

        if await self._remove_prev_message():
            await self.bot.send_message(chat_id=self.chat_id,
                                        text=self.bot_text,
                                        reply_markup=self.bot_markup)
        return None
